/*
 * fire.c
 *
 * fire at sectors or ships from forts, ships, and artillery units
 *
 * Ken Stevens, 1996
 */

#include "player.h"
#include "sect.h"
#include "ship.h"
#include "shell.h"
#include "file.h"
#include "treaty.h"
#include "optlist.h"

fire()
{
	int	numtimes = 1;	/* the number of times to fire */
	int	numcount;	/* number of times we've fired so far */
	int	attacker_ef_type = EF_BAD; /* attacker's ef type */
	int	trigger;
	struct 	sctstr vsect;	/* victim sector */	
	struct 	shpstr vship;	/* victim ship */	
	struct	genitem *vitem;
	struct	shelist vlist;	/* victim sector/ship wrapper */
	struct	shelist vl;	/* victim sector/ship */
	struct	shelist alist;	/* attacker list of shelling items */
	struct	shelist dlist;	/* list of defending items */
	natid	victim;		/* owner of victim sector/ship */
	s_char	*p;
	s_char	buf[1024];
	
	shell_init_shelist(&vlist);
	shell_init_shelist(&alist);
	shell_init_shelist(&dlist);

	/*
	 * Parse arguments
	 */

	if (*player->argp) {
		switch(**player->argp) {
		case 'f':
			trigger = SHELL_FIRE_COMMAND; break;
		case 't':
			trigger = SHELL_TORP_COMMAND; break;
		default:
			return RET_SYN;  /* impossible */
		}
	} else {
		return RET_SYN;  /* impossible */
	}

	if (player->argp[4] && *player->argp[4]) {
		attacker_ef_type = ef_byname(player->argp[4]);
		if (attacker_ef_type != EF_SECTOR &&
		    attacker_ef_type != EF_SHIP &&
		    attacker_ef_type != EF_LAND)
			return RET_SYN;
		if (!opt_FORTFIRE && attacker_ef_type == EF_SECTOR) {
			pr("Forts do not fire in this game\n");
			return RET_SYN;
		}
	}
	
	if (!(p = getstarg(player->argp[1], "Firing from? ", buf)) || !*p)
		return RET_SYN;
	
	if (!(p = getstarg(player->argp[2], "Firing at? ", buf)) || !*p)
		return RET_SYN;

	if (issector(p)) {
		vitem = (struct genitem *)&vsect;
		vitem->ef_type = EF_SECTOR;
		if (!sarg_xy(p, &vitem->x, &vitem->y) ||
		    !getsect(vitem->x, vitem->y, vitem)) {
			pr("No such sector exists!\n");
			return RET_OK;
		}
		insque(shell_create_item(vitem, 0, &vl, trigger),
		       &vlist.queue);
		if (!shell_valid_target(vitem, player->cnum, 1))
			return RET_OK;
		if (!trechk(player->cnum, vitem->own, LANFIR))
			return RET_OK;
#ifdef SLOW_WAR
		if (vitem->own &&
		    player->cnum != vitem->own &&
		    vsect.sct_oldown != player->cnum &&
		    getrel(getnatp(player->cnum), vitem->own) != AT_WAR) {
			pr("You're not at war with them!\n");
			return RET_OK;
		}
#endif /* SLOW_WAR */		
	} else {
		vitem = (struct genitem *)&vship;
		vitem->ef_type = EF_SHIP;
		if ((vitem->uid = atoi(p)) < 0 ||
		    !getship(vitem->uid, vitem) ||
		    !vitem->own) {
			pr("No such ship exists!\n");
			return RET_OK;
		}
		insque(shell_create_item(vitem, 0, &vl, trigger), &vlist.queue);
		if (!shell_valid_target(vitem, player->cnum, 1))
			return RET_OK;
		if (!trechk(player->cnum, vitem->own, SEAFIR))
			return RET_OK;
	}

	if (player->argp[3] && *player->argp[3]) {
		numtimes = atoi(player->argp[3]);
		if (numtimes > 20)
			numtimes = 20;
		if (numtimes < 1)
			return RET_SYN;
	}

	/*
	 * Get the list of items to shell from
	 */

	victim = vitem->own;

	shell_find_fire(&vlist, victim, player->argp[1], attacker_ef_type, &alist, player->cnum, trigger, 1);
	if (QEMPTY(&alist.queue)) {
		pr("Nothing to fire with...\n");
		return RET_OK;
	}

	/*
	 * Get the list of items which will be returning fire
	 */

	if (victim &&
	    victim != player->cnum)
		shell_find_fire(&alist, player->cnum, "*", EF_BAD, &dlist, victim, SHELL_RETURN_FIRE, 0);

	/*
	 * Fire the shells
	 */

	player->quiet_damage = 1; /* temprary hack for lib/common/damage.c */

	for (numcount=0; numcount < numtimes && !QEMPTY(&alist.queue); ++numcount) {

		if (!shell_valid_target(vitem, player->cnum, 1))
			break;

		/* attacker fires */

		shell_select_targets(&vlist, &alist, 1);
		if (QEMPTY(&alist.queue))
			break;
		shell_fire(&vlist, &alist);
		shell_damage(&vlist);

		/* defender fires */

		shell_select_targets(&alist, &dlist, 0);
		shell_fire(&alist, &dlist);
		shell_damage(&alist);

		shell_report_damage_delivered(&vlist, &alist, player->cnum, vitem, victim, trigger);
		shell_report_damage_received(&vlist, &alist, player->cnum);
		shell_report_damage_delivered(&alist, &dlist, victim, 0, player->cnum, SHELL_RETURN_FIRE);

		if (numtimes > 1) {
			vl.damage_received = 0.0;
			shell_reset(&vlist, &alist, 1);
			shell_reset(&alist, &dlist, 0);
		}
	}
	player->quiet_damage = 0;

	shell_free_list(&alist);
	shell_free_list(&dlist);

	if (vitem->ef_type == EF_SHIP && !vitem->own)
		pr("%s sunk!\n", prship((struct shpstr *)vitem));

	return RET_OK;
}

