#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "misc.h"
#include "match.h"
#include "file.h"
#include "struct.h"
#include "edbproc.h"
#include "nsc.h"
#include "genitem.h"

#define E_ALL -1
#define E_GROUP -2
#define E_AREA  -3
#define E_XY -4

struct castr *e_cadef; 

Empire_Query *Search(Empire_Database *, int, int, int, struct range, 
			  struct nscstr *,int);
			  

void *
edb_first_item(Empire_Query *q)
{

if(!q) return NULL;
q->cursor=0;
return (void *)q->results[0];

}

void edb_free_query(Empire_Query *eq)
{

  int n;

  if(!eq) return;

  free(eq->results);
  free(eq);

}

void *
edb_next_item(Empire_Query *q)
{

if(!q) return NULL;

if(q->cursor > q->nresults)
   return NULL;

return (void *)q->results[q->cursor++];

}

void 
AddToQueryList(Empire_Query *q,void *data)
{

q->results[q->nresults]=(struct genobject *)data;
q->nresults++;


}

int
Edb_typematch(s_char *buf, int type)
{
	int n;
	int	len;

	len = strlen(buf);
	switch (type) {
	case EF_SECTOR: {
		struct dchrstr *dcp;
	
		if (!buf[0] || buf[1])
			return -1;
		for (dcp=dchr,n=0; dcp->d_name; n++,dcp++)
			if (dcp->d_mnem == *buf)
				return n;
		}
		break;
	case EF_SHIP: {
		struct mchrstr *mcp;
	
		for (mcp=mchr,n=0; *mcp->m_name; n++,mcp++)
			if (strncmp(mcp->m_name, buf, len) == 0)
				return n;
		}
		break;
	case EF_LAND: {
		struct lchrstr *lcp;

		for (lcp=lchr,n=0; *lcp->l_name; n++,lcp++)
			if (strncmp(lcp->l_name, buf, len) == 0)
				return n;
		}
		break;
	case EF_PLANE: {
		struct plchrstr *pcp;
	
		for (pcp=plchr,n=0; *pcp->pl_name; n++,pcp++)
			if (strncmp(pcp->pl_name, buf, len) == 0)
				return n;
		}
		break;
	case EF_NUKE: {
		struct nchrstr *ncp;
	
		for (ncp=nchr,n=0; *ncp->n_name; n++,ncp++)
			if (strncmp(ncp->n_name, buf, len) == 0)
				return n;
		}
		break;
	default:
		break;
	}
	return -1;
}

int
Edb_stmtch(s_char *obj, s_char *base, int off, int size)
{
	s_char *str;
	int stat2;
	int i;
	int n;

	stat2 = M_NOTFOUND;
	str = base + off;
	for (i=0; *(s_char **)str; i++, str += size) {
		if ((n = mineq(obj, *(s_char **)str)) == ME_MISMATCH)
			continue;
		if (n == ME_EXACT)
			return i;
		if (stat2 != M_NOTFOUND)
			return M_NOTUNIQUE;
		stat2 = i;
	}
	return stat2;
}

int
Edb_nstr_exec(struct nscstr *conds, int ncond, caddr_t ptr, int type)
{
	struct nscstr *nsc;
	int op;
	int lhs;
	int rhs;

	for (nsc=conds; --ncond >= 0; nsc++) {
		lhs = Edb_decode(nsc->fld1, ptr, type);
		rhs = Edb_decode(nsc->fld2, ptr, type);
		op = nsc->oper;
		if ((op == '<' && lhs >= rhs) ||
		    (op == '=' && lhs != rhs) ||
		    (op == '>' && lhs <= rhs) ||
		    (op == '#' && lhs == rhs))
			return 0;
	}
	return 1;
}

int
Edb_decode(long code, s_char *addr, int type)

{
	int val;
	int nsc_code;
	struct natstr *np;

	val = (code & ~NSC_MASK) & 0xffff;
	
	/* handle negative numbers properly */
	/* this assumes a binary two's complement number representation */
	if (val>=0x8000) val -= 0x10000;
	
	nsc_code = code & NSC_CMASK;
	if (nsc_code == NSC_VAR) {
		val = getvar(val, addr, type); 
	} else if (nsc_code == NSC_OFF) {
		/*
		 * add offset to value
		 */
		addr += val;
		switch (code & NSC_TMASK) {
		case NSC_CHAR:
			val = *((s_char *) addr);
			break;
		case NSC_UCHAR:
			val = (int) *((unsigned char *) addr);
			break;
		case NSC_SHORT:
		case NSC_XCOORD:
		case NSC_YCOORD:
			val = *((short *) addr);
			break;
		case NSC_USHORT:
			val = *((u_short *) addr);
			break;
		case NSC_INT:
			val = *((int *) addr);
			break;
		case NSC_LONG:
			val = *((long *) addr);
			break;
		default:
			printf("bad type in decode: %x!\n",
				code & NSC_TMASK);
			val = 0;
			break;
		}
	}
	if (code & NSC_ROUND)
		val = round(val, 10);
	return val;
}


s_char *
Edb_nstr_comp(struct nscstr *np, int *size, int type, s_char *str)
{
	s_char *bp;
	s_char *cp;
	int c;
	s_char	ident[80];
	s_char	arg[255];

	strncpy(arg, str, sizeof(arg)-1);
	arg[sizeof(arg)-1] = 0;
	cp = arg;
	bp = ident;
	while ((c = *cp++) && bp < &ident[sizeof(ident)-1]) {
		if (c == '<' || c == '=' || c == '>' || c == '#')
			break;
		*bp++ = c;
	}
	*bp = 0;
	if (c == 0) {
		printf("'%s'? -- meaningless condition?\n", arg);
		return 0;
	}
	np[*size].oper = c;
	if (Edb_encode(ident, &np[*size].fld1, type) < 0)
		return 0;
	bp = ident;
	while ((c = *cp++) && bp < &ident[sizeof(ident)-1]) {
		if (c == '&')
			break;
		*bp++ = c;
	}
	*bp = 0;
	if (Edb_encode(ident, &np[*size].fld2, type) < 0)
		return 0;
	if (c == 0)
		cp--;
	(*size)++;
	return str + (cp - arg);
}

int
Edb_encode(s_char *str, long *val, int type)
{
	int i;


	if (str == 0) {
		*val = 0;
		return 0;
	}
	if (isdigit(*str) || *str == '-' && isdigit(str[1])) {
		/* XXX silently truncate to 16 bit int */
		*val = atoi(str) & 0xffff;
		return 1;
	}
	if ((i = Edb_typematch(str, type)) >= 0) {
		*val = i;
		return 1;
	}
	if (e_cadef != 0) {
		i = Edb_stmtch(str, (caddr_t)e_cadef, fldoff(castr, ca_name),
			sizeof(struct castr));
		if (i >= 0) {
			*val = e_cadef[i].ca_code|NSC_OFF;
			*val &= ~NSC_ROUND;
			return 1; 
		}
		if (i == M_NOTUNIQUE) {
			printf("%s -- ambiguous type selector\n", str);
			return 0;
		}
	}
	/*
	 * Only check for commodity selectors on objects which
	 * are allowed to have commodities.
	 */
	if (ef_flags(type) & EFF_COM) {
		i = Edb_stmtch(str, (caddr_t)var_ca, fldoff(castr, ca_name),
			sizeof(struct castr));
		if (i >= 0) {
			*val = var_ca[i].ca_code & ~NSC_ROUND;
			return 1;
		}
		if (i == M_NOTUNIQUE) {
			printf("%s -- ambiguous commodity selector\n", str);
			return 0;
		}
	}
	printf("%s -- not a valid selector\n", str);
	return 0;
}


Empire_Query *
edb_query(Empire_Database *edb,char *string)
{
  char            *ptr,*ptr2,*cp;
  char            qt[10],tmpbuf[20];
  struct nscstr   Cond[NS_NCOND]; 
  int             qtype,rnum=E_AREA,size,type;
  struct range    r;


/* get query type: land, ship, plane, nuke or sect */

  ptr=string;
  while(*ptr != (char) NULL && !isspace(*ptr)) ptr++;
  if ((ptr-string) > 8) {
    printf("Invalid string\n");
    return NULL;
  }
  strncpy(qt,string,ptr-string);
  qt[ptr-string+1]=0;

/* munch whitespace */  

  while(*ptr != (char) NULL && isspace(*ptr)) ptr++;

  if(strncmp(qt,"sect",4)==0) {
    qtype=V_Sector_Data;
    type=EF_SECTOR;
  }
  else if(strncmp(qt,"land",4)==0) {
    qtype=V_Land_Data;
    type=EF_LAND;
  }
  else if(strncmp(qt,"plan",4)==0){
    qtype=V_Plane_Data;
    type=EF_PLANE;
  }
  else if(strncmp(qt,"ship",4)==0) {
    qtype=V_Ship_Data;
    type=EF_SHIP;
  }
  else if(strncmp(qt,"nuke",4)==0) {
    qtype=V_Nuke_Data;
    type=EF_NUKE;
  }
  else 
    {
      printf("Invalid query request: Must be sect, land, ship or nuke\n");
      return NULL;
    }
  switch(qtype) {
  case V_Sector_Data: e_cadef=(struct castr *)&sect_ca; break;
  case V_Land_Data: e_cadef=(struct castr *)&land_ca; break;
  case V_Plane_Data: e_cadef=(struct castr *)&plane_ca; break;
  case V_Ship_Data: e_cadef=(struct castr *)&ship_ca; break;
  case V_Nuke_Data: e_cadef=(struct castr *)&nuke_ca; break;
  default: printf("Ack! invalid type.\n");
    return NULL;
  }

  if(!*ptr) {
    printf("Missing/invalid range...\n");
    return NULL;
  }

  if (*ptr=='*')  
    rnum=E_ALL;  /* look for all */
  else if (*ptr=='#') {
    if(!isdigit(*(++ptr)))
      {
	printf("Expect a number after #\n");
	return NULL;
      }
    rnum=atoi(ptr);
    if(rnum < 0 || rnum >= MAXNOR ) {
      printf("Out of range realm\n");
      return NULL;
    }
  } else if (isalpha(*ptr)) {
    r.lx=*ptr;
    rnum=E_GROUP;  /* look for in army/wing/fleet */
  } else {
    r.lx = atoi(ptr);

    while(*ptr != (char) NULL && (isdigit(*ptr) || *ptr == '-')) ptr++;

    if(*ptr==(char)NULL ||(*ptr != ':' && *ptr != ',')|| !isdigit(*(++ptr)) ) {
      printf("Bad area syntax, expect x1:x2,y1:y2 \n");
      return NULL;
    }

    if(*(ptr-1) == ',') {  /* x,y case */
      rnum=E_XY;
      r.ly=atoi(ptr);
    } else {   /* x:x1,y:y1 case */
      
      r.hx = atoi(ptr);
      
      
      while(*ptr != (char) NULL && (isdigit(*ptr) || *ptr == '-')) ptr++;
      
      if(*ptr == (char) NULL || *ptr != ',' || !isdigit(*(++ptr)) ) {
	printf("Bad area syntax, expect x1:x2,y1:y2 \n");
	return NULL;
      }
      r.ly = atoi(ptr);
      
      while(*ptr != (char) NULL && (isdigit(*ptr) || *ptr == '-')) ptr++;
      
      if(*ptr == (char )NULL || *ptr != ':' || !isdigit(*(++ptr)) ) {
	printf("Bad area syntax, expect x1:x2,y1:y2 \n");
	return NULL;
      }
      r.hy = atoi(ptr);
    }
  }
  
  while(*ptr != (char) NULL && *ptr != '?') ptr++;
  if((*ptr == (char) NULL))  /* No conditions */
    {
      return Search(edb,type,qtype,rnum,r,NULL,0); 
    }
  ptr++;
  cp = ptr;
  size=0;

  /*  Encode the list of conditions */

  while ((cp = Edb_nstr_comp(Cond, &size, type, cp)) && *cp)
    ;

  return Search(edb,type,qtype,rnum,r,(struct nscstr *)&Cond,size);

}

int
Edb_xyinrange(x, y, rp)
     coord   x;
     coord   y;
     struct  range *rp;
{
  if (rp->lx < rp->hx) {
    /* xrange doesn't wrap */
    if (x < rp->lx || x > rp->hx)
      return 0;
  } else {
    if (x < rp->lx && x > rp->hx)
      return 0;
  }
  if (rp->ly < rp->hy) {
    /* yrange doesn't wrap */
    if (y < rp->ly || y > rp->hy)
      return 0;
  } else {
    if (y < rp->ly && y > rp->hy)
      return 0;
  }
  return 1;
}

Empire_Query *
Search(Empire_Database *edb,int type, int qtype, int rnum, 
	    struct range r, struct nscstr *Clist, int size)
{

  
  int n, i, struct_size,sel=0;
  void *p;
  Empire_Query *query;
  Collection *c;
  Country *coun;

  if((query=malloc(sizeof(Empire_Query)))==NULL)
    return NULL;

  query->nresults=0;
  query->cursor=0;
  c = (Collection *) objectCollection(edb, qtype);

  if (!c) return;

  n = getCollectionCount(c);

  if((query->results=malloc(n*sizeof(struct genobject *)))==NULL)
    return NULL;

  memset(query->results,0,n*sizeof(struct genobject *));

  for (i=0;i<n;i++) {

    p =  getCollectionItem(c,i);
    if (!p) continue;

    switch(rnum) {
    case E_ALL: sel=1; break;
    case E_GROUP: 
      if( ((struct genitem *)p)->group == (char) r.lx)
	sel=1;
      break;
    case E_XY:
      if (((struct genitem *)p)->x == r.lx && ((struct genitem *)p)->y == r.ly)
	sel=1;
      break;
    case E_AREA:
      if(Edb_xyinrange(((struct genitem *)p)->x,((struct genitem *)p)->y,&r))
	sel=1;
      break;
    default:
      coun = edb_getCountry(edb);
      if(!coun) return NULL;  
      r.lx=coun->nat_b[rnum].b_xl;
      r.hx=coun->nat_b[rnum].b_xh;
      r.ly=coun->nat_b[rnum].b_yl;
      r.hy=coun->nat_b[rnum].b_yh;
      if(Edb_xyinrange(((struct genitem *)p)->x,((struct genitem *)p)->y,&r))
	sel=1;
      break;
    }
   if(sel) 
      {
	if(Clist==NULL||size==0)
	  AddToQueryList(query,p);
	else if (Edb_nstr_exec(Clist, size, p, type))
	  AddToQueryList(query,p);
	sel=0;
      }
  }

  query->cursor=0;
  return query;

}


